/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file C1RectSet.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

/* Author: Daniel Wilczak 2002-2004 */

#ifndef _CAPD_DYNSET_C1RECTSET_H_
#define _CAPD_DYNSET_C1RECTSET_H_

#include "capd/dynset/C1Set.h"
#include "capd/vectalg/Norm.h"
#include "capd/dynsys/C1DynSys.h"
#include "capd/geomset/CenteredAffineSet.h"
#include "capd/geomset/MatrixAffineSet.h"

/*
 C^1-Lohner algorithm.
 Derivative of the flow moved via QR decomposition (third method)
 the set part - QR decomposition (3-rd method)
 */

namespace capd {
namespace dynset {

template<typename MatrixT>
class C1RectSet: public C1Set<MatrixT> ,
                 public capd::geomset::CenteredAffineSet<MatrixT>,
                 public capd::geomset::MatrixAffineSet<MatrixT> {
public:
  typedef MatrixT MatrixType;
typedef  typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef capd::vectalg::Norm<VectorType,MatrixType> NormType;
  typedef C1Set<MatrixT> SetType;
  typedef capd::geomset::CenteredAffineSet<MatrixT> C0BaseSet;
  typedef capd::geomset::MatrixAffineSet<MatrixT> C1BaseSet;

  C1RectSet(int dim, const NormType& aNorm);
  C1RectSet(const NormType& aNorm, int dim);
  C1RectSet(const VectorType& x, const NormType& aNorm);
  C1RectSet(const VectorType& x, const VectorType& r, const NormType& aNorm);
  C1RectSet(const VectorType& x, const MatrixType& B, const VectorType& r, const NormType& aNorm);
  C1RectSet(const C1RectSet& set);
  ~C1RectSet() {delete m_Norm;}
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys);
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys, C1RectSet& result) const;

  std::string show(void) const;
  std::string name() const { return "C1RectSet"; }

  const NormType *getNorm(void) const;

  //using C0BaseSet::operator VectorType;
  operator VectorType() const{
    return C0BaseSet::operator VectorType();
  }
  operator MatrixType() const{
    return C1BaseSet::operator MatrixType();
  }
  C1RectSet& operator=(const VectorType &v);
  C1RectSet& operator=(const C1RectSet &S);

protected:
  using C0BaseSet::m_x; // set is represented as x + B*r in C0BaseSet
  using C0BaseSet::m_B;
  using C0BaseSet::m_r;
  using C1BaseSet::m_D; // derivatives are stored as D + Bjac * R in C1BaseSet
  using C1BaseSet::m_Bjac;
  using C1BaseSet::m_R;

  NormType *m_Norm; //the logarithmic norm used in the computation of a rough
  // enclosure for the variational part
};

// ------------------------------------------------

template<typename MatrixType>
inline void C1RectSet<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType> &c1dynsys){
  move(c1dynsys,*this);
}

template<typename MatrixType>
inline const typename C1RectSet<MatrixType>::NormType *C1RectSet<MatrixType>::getNorm(void) const
{
  return m_Norm;
}

}} // the end of the namespace capd::dynset

#endif // _CAPD_DYNSET_C1RECTSET_H_
/// @}
