/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Node_powsqr.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#ifndef _CAPD_MAP_NODE_POWSQR_HPP_
#define _CAPD_MAP_NODE_POWSQR_HPP_

#include "capd/capd/power.h"
#include "capd/interval/Interval.hpp"

namespace capd{
namespace map{


// ------------------------ operators -----------------------------------

template<class ScalarType>
ScalarType& PowNode<ScalarType>::operator()(int i)
{
  if(i<=this->m_maxComputedDerivative)
    return this->value[i];
  this->m_maxComputedDerivative = i;

  (*(this->left))(i);
  (*(this->right))(i);

  ScalarType c=this->right->value[0];
  int cInt = toInt(c);
  if (c==ScalarType(1.))
    return this->value[i] = this->left->value[i];

  if (i==0)
  {
    if(ScalarType(cInt) == c)
        return *(this->value) = power(*(this->left->value),cInt);
    ScalarType temp = c * log(*(this->left->value));
    return *(this->value) = exp(temp);
  }

  ScalarType result(0.);

  if(this->left->value[0]==0)
  {
    ScalarType *temp1 = new ScalarType[i+1];
    ScalarType *temp2 = new ScalarType[i+1];

    int j,k,p;
    for(j=0;j<=i;++j)
      temp1[j]=this->left->value[j];

    for(k=1;k<c;++k)
    {
      for(p=0;p<=i;++p)
      {
        temp2[p]=ScalarType(0.);
        for(j=0;j<=p;++j)
          temp2[p] += this->left->value[j]*temp1[p-j];
      }
      for(p=0;p<=i;++p)
        temp1[p]=temp2[p];
   }

   result = temp1[i];
   delete []temp1;
   delete []temp2;
   return this->value[i] = result;
  }

  for(int j=0;j<i;++j)
    result+= ScalarType(i*c-j*(c+1)) * this->value[j] * this->left->value[i-j];

  return this->value[i] = result / (ScalarType(i) * this->left->value[0]);
}

// specialisation for intervals only
template < typename T_Bound, typename T_Rnd>
typename PowNode< capd::intervals::Interval<T_Bound,T_Rnd> >::ScalarType&
  PowNode< capd::intervals::Interval<T_Bound,T_Rnd> >::operator()(int i)
{

  if(i<=this->m_maxComputedDerivative)
    return this->value[i];
  this->m_maxComputedDerivative = i;


  (*(this->left))(i);
  (*(this->right))(i);
  int c = toInt((this->right->value[0]).leftBound());

  if(this->right->value[0]==ScalarType(1.))
    return this->value[i] = this->left->value[i];

  if (i==0)
    return this->value[i] = (
        ScalarType(c) == this->right->value[i] ?
          power(this->left->value[i],c) : power(this->left->value[i],this->right->value[i])
        );

  ScalarType result(0.);

  if(!(this->left->value[0].contains(0.0)))
  {
    for(int j=0;j<i;++j)
      result+= (this->right->value[0]*ScalarType(i)-(this->right->value[0]+1)*ScalarType(j))
                * this->value[j] * this->left->value[i-j];
    return this->value[i] = result/(ScalarType(i) * this->left->value[0]);
  }

  if(c<=0)
  {
    std::ostringstream bufor;
    bufor << "function error: can not compute Taylor series for x^(" << c << ") if x contains zero";
    throw(std::runtime_error(bufor.str()));
  }

  ScalarType *temp1 = new ScalarType[i+1];
  ScalarType *temp2 = new ScalarType[i+1];
  int j,p;
  int k;
  for(j=0;j<=i;++j)
    temp1[j]=this->left->value[j];

  for(k=1;k<c;++k)
  {
    for(p=0;p<=i;++p)
    {
      temp2[p] = ScalarType(0.);
      for(j=0;j<=p;++j)
        temp2[p] += this->left->value[j]*temp1[p-j];
    }
    for(p=0;p<=i;++p)
      temp1[p]=temp2[p];
  }

  result = temp1[i];
  delete []temp1;
  delete []temp2;

  return this->value[i] = result;
}


// ----------------------------------------- SqrNode operator ----------------------------

template<typename ScalarType>
ScalarType& SqrNode<ScalarType>::operator()(int i)
{
  if(i<=this->m_maxComputedDerivative)
    return this->value[i];
  this->m_maxComputedDerivative = i;


  (*(this->left))(i);

  ScalarType result(0.);
  int k = (i >> 1);
  if(i%2) // in i is odd  we have result = sum_{j=0}^k 2*p[j]*p[i-j]
  {
    for(int j=0;j<=k;++j)
      result += ScalarType(2)*this->left->value[j]*this->left->value[i-j];
  }
  else
  {
    for(int j=0;j<k;++j)
      result += ScalarType(2)*this->left->value[j]*this->left->value[i-j];
    result += power(this->left->value[k],2);
  }

  return this->value[i] = result;
}

// ----------------------------------------- SqrtNode operator ----------------------------

template<typename ScalarType>
ScalarType& SqrtNode<ScalarType>::operator()(int i)
{
  if(i<=this->m_maxComputedDerivative)
    return this->value[i];
  this->m_maxComputedDerivative = i;

  (*(this->left))(i);
  if(!i)
  {
    if(!(this->left->value[0]>0))
    {
      std::ostringstream bufor;
      bufor << "function error: ScalarType sqrt error: nonpositive argument " << this->left->value[0];
      throw (std::runtime_error(bufor.str()));
    }
    return this->value[i] = sqrt(this->left->value[0]);
  }

  ScalarType result(0);
  for(int j=0;j<i;++j)
    result += ScalarType(0.5*(i-3*j)) * this->value[j] * this->left->value[i-j];

  return this->value[i] = result / (ScalarType(i) * this->left->value[0]);
}

}} // namepsace capd::map

#endif // _CAPD_MAP_NODE_POWSQR_HPP_

/// @}
