/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file  PoincareMap_templateMembers.h
///
/// @author Daniel Wilczak
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#ifndef _CAPD_POINCARE_POINCARE_MAP_TEMPLATE_MEMBERS_HPP_
#define _CAPD_POINCARE_POINCARE_MAP_TEMPLATE_MEMBERS_HPP_

#include <cassert>
#include "capd/poincare/PoincareMap.h"
#include "capd/poincare/BasicPoincareMap.hpp"
#include "capd/poincare/timeStep.hpp"

namespace capd{
namespace poincare{

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet)
{
// The originalSet after this function contains a set just ater section.
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
// We move the set to be very close to the section   
   T setAfterTheSection = reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet, setAfterTheSection);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, MatrixType& der)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
   T setAfterTheSection = reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet, setAfterTheSection);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, MatrixType& der, MatrixType hessian[])
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = hessian;
   partialDerivatives = NULL;
   T setAfterTheSection = reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet, setAfterTheSection);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, CnCoeffType& result)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = &result;
   T setAfterTheSection = reachSection<T>(originalSet);

// and compute the PoincareMap image
   return result()=crossSection<T>(originalSet, setAfterTheSection);
}

// -------------------- protected functions ---------------------------------

// this function is just for compatibility with C^1 and C^2 routines
// see the version for C^1-set types below
template <typename DS, typename FunT>
inline void PoincareMap<DS, FunT>::computeJacEnclosure(C0Set& S, VectorType& V)
{
   
}

/*__________________________________________________________________________*/
/// Function returns section sign. It also checks If we crossed section 
/// and returned in one step. In this case it throws an exception.
/// @param[in]      theSet      the set after making a step,
/// @param[in, out] position    position before time step,
/// @param[in]      updatePosition   flag that decided if position has to be updated
/// @param[in]      bound       rough enclosure for trajectory during whole time step
/// @return    section sign 
template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::ScalarType PoincareMap<DS, FunT>::getSign(
  const T & theSet,  VectorType & position, bool updatePosition, const VectorType & bound 
){
  // we check if we cross the section and than return during one step
  // i.e. if during section crossing the section gradient is orthogonal to vector field 
  checkTransversability(theSet, bound);
  
  // Now we are sure that either sing is constant during the time step 
  // or we crossed the section but tranversally, so we can compute sign on position after the step.
  if(updatePosition){
    position = VectorType(theSet);
    return section(position);
  }
  else
    return section(VectorType(theSet));
}
/*__________________________________________________________________________*/
/// Function returns section sign. It also checks If we crossed section 
/// and returned in one step. In this case it throws an exception.
/// @param[in]      theSet      the set after making a step,
/// @param[in, out] position    position before time step,
/// @param[in]      updatePosition   flag that decided if position has to be updated
/// @return    section sign 
template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::ScalarType PoincareMap<DS, FunT>::getSign(
  const T & theSet,  VectorType & position, bool updatePosition
){
  return getSign(theSet, position, updatePosition, dynamicalSystem.enclosure(position)); 
}
/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
T PoincareMap<DS, FunT>::reachSection(T& theSet)
// The common procedure to reach section for all types of the sets.
// As a result we obtain the set closer to section than sectionFactor * sizeOfSet
{
  ScalarType step = dynamicalSystem.getStep(); // time step used to reach the section
  
  //------------------------------------------------------------------------------------------------
  // LEAVING SECTION AND GOING UP TO THE POINT WHERE SECTION HAS CORRECT SIGN
  // We want to leave section and reach point where the sign of the section function,
  // according to crossingDirection, indicate that next section crossing will be in a good direction.
  
  VectorType setPosition = VectorType(theSet);
  ScalarType sign = section(setPosition);
  
  while( (sign.contains(0.0)) ||                  // We are on the section
         !((crossingDirection == None) ||         // section sign is not correct
           ((crossingDirection == MinusPlus) && (sign < 0.0)) ||
           ((crossingDirection == PlusMinus) && (sign > 0.0))
          )
  ){
    theSet.move(dynamicalSystem);   // we make one step to leave the section and try again
    sign = getSign(theSet, setPosition, true);
    if(stepControl)
    {
      ScalarType newStep = computeNextTimeStep(dynamicalSystem,setPosition);
      if(step<0)
        newStep = -newStep;
      dynamicalSystem.setStep(newStep);
    }
  }

//------------------------------------------------
// first return to section with constant time step
  
   assert(!sign.contains(ScalarType(0.0)));
   sign = sign.mid();            // we save the sign of the section function, sign do not contain zero
   ScalarType check = sign;      // used for checking if we cross the section (if sign is changed)
   T temp(theSet);
   T *original = &theSet, *tempSet = &temp;

// now we try to move the set closer to section than one current time step
   while(check*sign>0)
   {
      original->move(dynamicalSystem,*tempSet);
      check = getSign(*tempSet, setPosition, true);
      if(stepControl)
      {
        ScalarType newStep = computeNextTimeStep(dynamicalSystem,setPosition);
        if(step<0)
          newStep = -newStep;
        dynamicalSystem.setStep(newStep);
      }
      std::swap(original, tempSet);
  } // end while
   
   
   T setAfterTheSection = *original;
   
   // We take the last set with not changed section sign
   if(&theSet != tempSet) theSet = *tempSet;
   
//==================================================================================================
// We want to come to section closer than  sectionFactor * size of set in section gradient direction

   setPosition = VectorType(theSet);
   VectorType Grad = section.gradient(setPosition);

   // maximal distance is sectionFactor * size of the set in the direction parpendicular to the section 
   typename ScalarType::BoundType maxDistance = diam(abs(setPosition*Grad)).rightBound()*sectionFactor;
   int maxIterations = 30; // usually 2-3 iterations are necessary to come close enoungh to the section
  
   original = &theSet;
   tempSet = &temp;
  
   while(maxIterations)
   {
      ScalarType approxDistance = (abs(section(setPosition))).left();
      if(approxDistance.leftBound() < maxDistance)      // we are close enough
         break;
      
      VectorType setCenter = midVector(setPosition);
      VectorType velocity = dynamicalSystem.getField()(setCenter);
      ScalarType velocityInNormalDirection = abs(velocity * Grad);
      // We want to stop at distance equal to 1/15 of maxDistance 
      // (1/15 is an experimental value, chosen as a good working) 
      ScalarType approxTime = ((approxDistance-(maxDistance/15))/ velocityInNormalDirection).mid().left();
      
      if (step<0)
         approxTime = -approxTime;

      dynamicalSystem.setStep(approxTime);
      original->move(dynamicalSystem,*tempSet);
   
      int correctionTries = maxCorrectionTries ;
      while(true)
      {  
         check = getSign(*tempSet, setPosition, false);
        
         if((!(check*sign>0)) && (correctionTries-- != 0))
         {  // an aproximated time step is to big, we touched or crossed the section 
            // so we shorten time step multiplying it by some correction factor from (0,1)
            ScalarType step = dynamicalSystem.getStep();
            ScalarType newStep = step * timeStepCorrectionFactor;
            if(step ==newStep)
              break;
            dynamicalSystem.setStep(newStep);
            original->move(dynamicalSystem,*tempSet);
         }        
         else 
           break;
      } // end while

      if(correctionTries != -1) {
        std::swap(original, tempSet);
        setPosition = VectorType(*original);
        Grad = section.gradient(setPosition);
        maxIterations--;
      }else{
        break;
      }
   }

   dynamicalSystem.setStep(step);
   if(&theSet != original) 
     theSet = *original;
   
   return setAfterTheSection;
}

/////////////////////////////////////////////////////////////////////////////////////
///
/// Function used to cross the section for all types of sets
///
/// @param[in,out] theSet              set just before the section, on exit is equal to setAfterTheSection
/// @param[in]     setAfterTheSection  set on the section or just after it
/// @returns        set that contains value of Poincare map for given set theSet

template <typename DS, typename FunT> 
template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::crossSection(
   T& theSet, const T & setAfterTheSection
){
// theSet is very close to the section
// we compute an approximate time needed to cross the section
// if the approximate values of time step is to small to change the sign of the section function
// the 'smallTime' time is used to move the set
   ScalarType step = dynamicalSystem.getStep();
   int order = dynamicalSystem.getOrder();
   if(sectionOrder)
   {
    dynamicalSystem.setOrder(sectionOrder);
   }
   // We want stop after section, closer than  maxDistance =  sectionFactor * size of set in section gradient direction 
  VectorType setPosition = VectorType(theSet);
  VectorType Grad = section.gradient(setPosition);
  typename ScalarType::BoundType maxDistance = diam(abs(setPosition*Grad)).rightBound()*sectionFactor;
  typename ScalarType::BoundType approxDistance = - abs(section(setPosition)).rightBound(); 
  
  VectorType result = setPosition;
  
  ScalarType check = section(setPosition);
  if(check.contains(0.0)){
    throw PoincareException<T>("Poincare exception in section crossing : initial set is already on the section ", theSet, check );
  }
  ScalarType sign = mid(check);
  
  
  // direction of section crossing =  PlusMinus or MinusPlus
  // we check sign of the section instead of checking value of variable crossingDirection because it can be set to None 
  typename BasicPoincareMap<DS, FunT>::CrossingDirection direction = ( check > 0)? PlusMinus: MinusPlus;
  
  T temp(theSet); // we save the set before section, used in throwing exception only
  saveMatrices(theSet); // we save the \frac{d}{dx}\phi

  T* original = &theSet;
  T* tempSet = &temp;
  bool sectionReached = false;
  
  while(!(check*sign < 0)){
    
    VectorType fieldDirection = dynamicalSystem.getField()(midVector(setPosition));
    ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
    ScalarType approxTime = ((maxDistance/2 - approxDistance ) / fieldSectionDirection).mid();
    
    // if approxTime is too big, it can cause problems with finding an enclosure.
    // In this case we use original time step (possibly several times)
    approxTime = capd::min(approxTime, abs(step));
    
    /// approxTime cannot be too small, otherwise it can occur that we will never cross the section
    approxTime = capd::max(approxTime/10, minCrossingTimeStep);
    if (step<0)
      approxTime = -approxTime;
   
    int correctionTries = maxCorrectionTries;
    while(true){
      dynamicalSystem.setStep(approxTime.left());
      original->move(dynamicalSystem,*tempSet);  
      setPosition = VectorType(*tempSet);
      check = section(setPosition);
      
      if(direction == PlusMinus)
        approxDistance = - check.rightBound();
      else
        approxDistance = check.leftBound();
      
      if((approxDistance  > maxDistance)&& (correctionTries-- != 0)){  // the step was too big
        approxTime *= timeStepCorrectionFactor;  // so we shorten it
      }else{
        break;
      }
    }
   
    VectorType bound = dynamicalSystem.enclosure(VectorType(*original));
    checkTransversability(*original, bound);
    if((!sectionReached) && (check*sign > 0)){ // we did not touch the section yet
       result = setPosition;
       saveMatrices(*tempSet);
    }else{
      
      computeJacEnclosure(theSet, bound);
      
      // TODO: It was computed already in checkTranversability 
      VectorType Fresult = dynamicalSystem.getField()(bound);

      for(int i=0;i<m_dim;i++){
        if(!(Fresult[i].contains(0.)))
           result[i] = intervalHull(result[i],setPosition[i]);
        else
           result[i] = intervalHull(result[i],bound[i]); // bound is an enclosure
      }
    }
    T* swapSet = original;
    original = tempSet;
    tempSet = swapSet;
    
  } // end while

  dynamicalSystem.setStep(step);
  theSet = setAfterTheSection;
  if(sectionOrder)
  {
    dynamicalSystem.setOrder(order);
  }
  return result;
}

/// Function checks if we crossed section and then returned in one step. 
/// In this case it throws an exception of PoincareException<T> type.
template <typename DS, typename FunT> template<typename T>
void PoincareMap<DS, FunT>::checkTransversability(
  const T & theSet, const VectorType & bound 
){
  ScalarType check = section(bound);
 
  if(subset(ScalarType(0.0), check)) {  // Is the section crossed?
    ScalarType innerProduct = section.gradient(bound)*dynamicalSystem.getField()(bound);
    if (innerProduct.contains(0.0)) {  // Is the vector field orthogonal to section gradient?
      throw PoincareException<T>(
              "Poincare error: possible nontranversal return to the section ", theSet, theSet,
              bound, dynamicalSystem.getField()(bound),
              check, section.gradient(bound), innerProduct
              );
    }
  }
  
}

}} // namespace capd::poincare

#endif // _CAPD_POINCARE_POINCARE_MAP_HPP_

/// @}
