/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file TimeMap.hpp
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _CAPD_POINCARE_TIME_MAP_HPP_
#define _CAPD_POINCARE_TIME_MAP_HPP_

#include "capd/poincare/TimeMap.h"

namespace capd{
namespace poincare{



template <typename DS>
typename TimeMap<DS>::VectorType TimeMap<DS>::operator()(ScalarType time, VectorType v)
// the point on trajectory just after time 'time' 
{
  ScalarType step = dynamicalSystem.getStep();
  ScalarType t(0.0), lastStep;

  while(t + step <= time){
    v = dynamicalSystem(v);
    t += step;
  }

  // we make last step
  lastStep = time - t;
  dynamicalSystem.setStep(lastStep);
  v = dynamicalSystem(v);

  dynamicalSystem.setStep(step);

  return v;
}

template <typename DS>
typename TimeMap<DS>::VectorType TimeMap<DS>::operator()(ScalarType time, VectorType v, MatrixType &der)
// the point on trajectory just after time 'time' 
{
  ScalarType step = dynamicalSystem.getStep();
  ScalarType t(0.0), lastStep;
  der = MatrixType::Identity(m_dim);
  MatrixType oneStepDerivative;

  while(t + step <= time){
    v = dynamicalSystem(v,oneStepDerivative);
    der = oneStepDerivative * der;
    t += step;
  }

  // we make last step
  lastStep = time - t;
  dynamicalSystem.setStep(lastStep);
  v = dynamicalSystem(v, oneStepDerivative);
  der = oneStepDerivative * der;

  dynamicalSystem.setStep(step);

  return v;
}

template <typename DS>
template <typename CnCoeffType>
typename TimeMap<DS>::VectorType TimeMap<DS>::operator()(ScalarType time, const VectorType& v, CnCoeffType& result)
// the point on trajectory just after time 'time' 
{
  result.clear();
  result() = v;
  result.setMatrix(MatrixType::Identity(v.dimension()));
  ScalarType step = dynamicalSystem.getStep();
  ScalarType t(0.0), lastStep;

  while(t + step <= time){
    dynamicalSystem(result);
    t += step;
  }

  // we make last step
  lastStep = time - t;
  dynamicalSystem.setStep(lastStep);
  dynamicalSystem(result);

  dynamicalSystem.setStep(step);

  return VectorType(result);
}

template<typename DS> template<typename SetType>
typename DS::VectorType TimeMap<DS>::operator()(ScalarType Time, SetType &originalSet)
{

  ScalarType step = dynamicalSystem.getStep();

  ScalarType t(0.0);

  while((t + step) < Time){
    originalSet.move(dynamicalSystem);
    t += step;
  }

   // we make last step
  ScalarType lastStep = Time - t;
  dynamicalSystem.setStep(lastStep);
  originalSet.move(dynamicalSystem);

  dynamicalSystem.setStep(step);
  return VectorType(originalSet);
}

template<typename DS> template<typename SetType>
typename DS::VectorType TimeMap<DS>::operator()(ScalarType time, SetType &originalSet, MatrixType & derivative)
{

  ScalarType step = dynamicalSystem.getStep();

  ScalarType t(0.0);

  while((t + step) < time){
    originalSet.move(dynamicalSystem);
    t += step;
  }

   // we make last step
  ScalarType lastStep = time - t;
  dynamicalSystem.setStep(lastStep);
  originalSet.move(dynamicalSystem);

  dynamicalSystem.setStep(step);
  derivative = (MatrixType) originalSet;
  return VectorType(originalSet);
}



}} // namespace capd::poincare


#endif // #define _CAPD_POINCARE_TIME_MAP_HPP_

/// @}
