#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>

#include "typedefs.h"

#include "capd/newton/Krawczyk.h"
#include "capd/newton/Newton.h"

#include "textfile.h"
#include "timeused.h"

#include "nbody.h"
#include "nBodyProblem.h"
#include "nBodyProblem2.h"

#include "toText.h"
#include "choreoExist.h"


template <typename MapType>
std::string  KrawczykInfo(capd::newton::Krawczyk<MapType> &k)
{
   std::ostringstream str;

   str    << "\n\n Initial Conditons for iteration " << k.numberOfIterations
          <<": \n\n x0 = " << k.x0
          << "\n\n X =  " << k.X
          << "\n\n max diam of X = "  << size(k.X).rightBound()<<"\n\n"
          << toText( "K(x0, X, F)", k.K, 16, true)
          << "\n\n other data from computations :\n"
          << toText("dF(X)", k.dF_X, 10, true)
          << toText("F(x0)", k.F_x0, 10, true) <<"\n\n";

   return str.str();
}


void ProveChoreographyExistence(FloatVector &x0, double size,
                                int numberOfBodies,  int nDiv,  Interval time,
                                int order,  ScalarType step,
                                int Xorder, ScalarType Xstep)
{

   sout << "\n " << numberOfBodies << " BODY PROBLEM \n";

   if (nDiv >1) sout<< "\n  number of division in parallel shooting: " << nDiv;

   sout << "\n parameters of computations : \n   F(x0) :  order = " << order << ", step = " << step
          << "\n   dF(X) :  order = " << Xorder << ", step = " << Xstep;


   NBodyProblem nbody(numberOfBodies, nDiv, time, order, step, Xorder, Xstep);

   capd::newton::Krawczyk<NBodyProblem> krawczyk(nbody) ;

   int code = krawczyk.proof(x0, size);
  
   sout << KrawczykInfo(krawczyk);

   switch(code){

      case 1: if(nbody.uniquenessTest())
              {
                 sout << "\n\n###############################################################";
                 sout << "\n\n THERE EXISTS CHOREOGRAPHY IN THE FULL SPACE. \n" ;
              }
              else
                 sout << "\n There exists zero in reduced space, but uniqueness test failed \n";
              break;

      case 0:  sout << "\n There are no choreographies in the given set X \n";
              break;

      case -1:
      case -2:  sout << "\n We cannot conclude. Maybe estimates are no good enough. \n";
              break;
   }
}

void ProveChoreographyExistence(FloatVector &x0, double size,
                                int numberOfBodies,  int nDiv,  Interval time,
                                int order1, int order2, int orderStep,
                                ScalarType step1, ScalarType step2, ScalarType stepFactor,
                                int Xorder1, int Xorder2, int XorderStep,
                                ScalarType Xstep1, ScalarType Xstep2, ScalarType XstepFactor)

{

    sout << "\n " << numberOfBodies << " BODY PROBLEM \n";

    if (nDiv >1) sout<< "\n  number of division in parallel shooting: " << nDiv;
 
    sout << "\n parameters of computations : \n   F(x0) :  order = " << order1 <<".."<<order2
        << ", step = " << step1<<"..."<<step2
        << "\n   dF(X) :  order = " << Xorder1<<"..."<< Xorder2
        << ", step = " << Xstep1<<"..."<<Xstep2;


   NBodyProblem2 nbody(numberOfBodies, nDiv, time,
                       order1, order2, orderStep,
                       step1,  step2,  stepFactor,
                       Xorder1, Xorder2, XorderStep,
                       Xstep1, Xstep2, XstepFactor);

   capd::newton::Krawczyk<NBodyProblem2> krawczyk(nbody) ;

   int code = krawczyk.proof(x0, size);

   sout << KrawczykInfo(krawczyk);

   switch(code){

      case 1: if(nbody.uniquenessTest())
              {
                 sout << "\n\n###############################################################";
                 sout << "\n\n THERE EXISTS CHOREOGRAPHY IN THE FULL SPACE. \n" ;
              }
              else
                 sout << "\n There exists zero in reduced space, but uniqueness test failed \n";
              break;

      case 0:  sout << "\n There are no choreographies in the given set X \n";
              break;

      case -1:
      case -2:  sout << "\n We cannot conclude. Maybe estimates are no good enough. \n";
              break;
   }
}



FloatVector loadData(const char * name, int numberOfBodies)
{

   std::ifstream file(name);

   if(!file)
     throw std::runtime_error("Cannot open file !");

   FloatVector x0(numberOfBodies*4-6),
               x(numberOfBodies*4);

   file >> x;

   x0[0]=x[0];
   for(int i = 1; i <numberOfBodies*4-6; ++i)
      x0[i] = x[i+2];

   return x0;

}

FloatVector loadData(const char * name, int numberOfBodies, int nDiv)
{

   std::ifstream file(name);

   if(!file)
     throw std::runtime_error("Cannot open file !");

   FloatVector x0((numberOfBodies-1)*nDiv*4-2),
               x(numberOfBodies*4*nDiv);

   file >> x;

   x0[0]=x[0];
   for(int i = 1; i <numberOfBodies*4-6; ++i)
      x0[i] = x[i+2];

   for(int j = 1; j<nDiv; ++j)
      for(int i = 0; i <numberOfBodies*4-4; ++i)
          x0[(numberOfBodies-1)*j*4 - 2 + i] = x[numberOfBodies*4*j+i];

   return x0;

}

FloatVector loadData(double x[], int numberOfBodies, int nDiv)
{
   FloatVector x0((numberOfBodies-1)*nDiv*4-2);

   x0[0]=x[0];
   for(int i = 1; i <numberOfBodies*4-6; ++i)
      x0[i] = x[i+2];

   for(int j = 1; j<nDiv; ++j)
      for(int i = 0; i <numberOfBodies*4-4; ++i)
          x0[(numberOfBodies-1)*j*4 - 2 + i] = x[numberOfBodies*4*j+i];

   return x0;
}

FloatVector loadData(FloatVector x, int numberOfBodies, int nDiv)
{
  FloatVector x0((numberOfBodies-1)*nDiv*4-2);

  x0[0]=x[0];
  for(int i = 1; i <numberOfBodies*4-6; ++i)
    x0[i] = x[i+2];

  for(int j = 1; j<nDiv; ++j)
    for(int i = 0; i <numberOfBodies*4-4; ++i)
      x0[(numberOfBodies-1)*j*4 - 2 + i] = x[numberOfBodies*4*j+i];

  return x0;
}

