/////////////////////////////////////////////////////////////////////////////
//   Package:          Choreographies in the N-Body Problem
/// @addtogroup Stability
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file 3stability.cpp
///
/// Performs proof of the stability of the figure eight solution.
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////
// Copyright (C) Tomasz Kapela 2006. 


#include <iostream>
#include <string>
#include <iomanip>
#include "textfile.h"
#include "timeused.h"

#include "typedefs.h"
#include "nBodyProblem.h"
#include "proof.h"
#include "toText.h"

#define M_2PI  6.28318530717958647692

using namespace capd;
///////////////////////////////////////////////////////////////////////////
// loadData
///
///  Function that loads trajectory data from file and reduces it 
///   by removing 2nd and 3rd coordinate
///
///  @remark In this version trajectory is fixed (the Eight) 
///          and data of trajectory is given explicitly.  
///
///////////////////////////////////////////////////////////////////////////
FloatVector loadData(const char * name)
{
   ScalarType xd[] = {1.076143733510925,   0.4682662184090647,
             -0.5380718667554619,  -0.3437068277582447, -1.099603752075198,   -0.2341331092045306};
   FloatVector x0(6, xd);

   return x0;
}

///////////////////////////////////////////////////////////////////////////
// trace 
///  
///  Computes Trace of a given matrix.
/// 
///  Trace is a sum of all elements on the main diagonal
///
///////////////////////////////////////////////////////////////////////////
template <typename MatrixType>
typename MatrixType::ScalarType trace(const MatrixType& A)
{
   typename MatrixType::ScalarType T(0.0);

   for(int i = 0; i<A.numberOfRows(); ++i)
      T += A[i][i];

   return T;
}
///////////////////////////////////////////////////////////////////////////
// secondTrace 
///  
///  It returns a sum of determinants of all 2x2 matrix 
///  from main diagonal
///
///////////////////////////////////////////////////////////////////////////
template <typename MatrixType>
typename MatrixType::ScalarType secondTrace(const MatrixType& A)
{
   typename MatrixType::ScalarType T(0.0);
   int dim = A.numberOfRows();

   for(int i = 0; i<dim; ++i)
      for(int j = i+1; j<dim; ++j)
        T += A[i][i]*A[j][j]-A[i][j]*A[j][i];

   return T;
}

///////////////////////////////////////////////////////////////////////////
//  stability
///
/// It checks eliptical stability of the Eight solution
///
/// From the Hamiltonian system theory  we know that monodromy matrix 
/// has at least 8 eigenvalues equal to 1.
/// We check that the remaining 4 eigenvalues are on unit circle and are not real. 
///
/// @param[in]  An interval matrix which contains monodromy matrix of the Eight solution
///
/// @returns  true if we prove linear stability
///           false othervise
///
///////////////////////////////////////////////////////////////////////////
template <typename MatrixType>
bool stability(const MatrixType &A)
{ 
   typename MatrixType::ScalarType
      a(1.0),
      b(4. - trace(A)) ,
      c(secondTrace(A) + 4 * b - 8) ;

   typename MatrixType::ScalarType 
      delta = (b*b) - 4*a*c;

   sout //<< "\nTrace(A) = " << trace(A) << "\n subTrace" << secondTrace(A)
        << "\n x1, x2 are solutions of ax^2 + bx + c = 0 where "
        << "\n a = " << a << "\n b= " << b << "\n c = " << c << "\n delta = " << delta;

   try{
     
     if(delta >=0.0)
     {   
       typename MatrixType::ScalarType
         x1 = (-b + sqrt(delta))/ (2*a),
         x2 = (-b - sqrt(delta))/ (2*a);
       sout << "\n\n x1 = " << x1 << " \n\n x2 = " << x2;
       if((x2 <0.0) && (x1 >0.0 ) && (x1 < 2.0))
         return true;
     }else{
       sout << "\n\n Cannot compute x-intercepts - deltha is not nonnegative.";
       return false;
     }
   }
   catch(std::exception &e){
     sout << "\n Exception: " <<  e.what() << "\n   Most probably deltha is not nonnegative. \n";
   }
   return false;
}



int main()
{
    program_time = 0 ;

   try{

     timeused timeMeter;
     scon.logfile("consoleOutput.txt");
     sout.logfile("raport3.txt");

     sout << " STABILITY OF THE EIGHT SOLUTION\n\n";

#if((MP == 1) || (MP == 2))
     capd::multiPrec::PrecisionType  prec = 100;
     capd::multiPrec::MpfrClass::setDefaultPrecision(prec);
     ScalarType size = 1.e-12;
     sout << "Precision used (in mantisa bits): " << capd::multiPrec::MpfrClass::getDefaultPrecision() <<"\n";
#else
     double size = 1.e-9;
#endif
           
     int numberOfBodies = 3;
     int nDiv = 1;
     Interval time = Interval(M_2PI/3./nDiv);

     NBodyProblem nbody(numberOfBodies, nDiv, time, 20, 1./128., 20, 1./128.);

     FloatVector x = loadData("3body.nbp");
     
     IntervalVector X(nbody.dimension()),
                    x0(nbody.dimension());
     
     for(int i=0; i< nbody.dimension(); i++)
     {
        X[i]=Interval(x[i]-size, x[i]+size);
        x0[i] = Interval(x[i]);
     }
     scon << "\n Computation of the Krawczyk operator. ";
     IntervalVector K = KrawczykOperator(x0, X, nbody);

     sout << toText("K", K) << "\n";

     bool unique = nbody.uniquenessTest();

     if((subsetInterior(K,X))&& unique) 
     {
         sout << "\n\n###############################################################";
         sout << "\n\n THERE EXISTS CHOREOGRAPHY IN THE FULL SPACE. \n" ;

         int howManyImprovements = 0;
         if(howManyImprovements > 0)
            scon << "\n We try to improve estimates by repeated computation of Krawczyk operator";
         for(int i=1; i<=howManyImprovements; ++i)
         {
            if(!subset(x0, K))
               x0 = midVector(K);
            K = KrawczykOperator(x0, K, nbody);
            sout << "\n improvement no. " << i << "\n" << toText("K", K);
         }


         IntervalMatrix dP = nbody.monodromyMatrix(K);
         sout <<"\n\n dP = " << dP;

         bool isStable = stability(dP);
                  
          if(isStable) 
            sout << "\n\n The proof of the Eight stability is done. \n";
         else
            sout << "\n\n We was not able to prove Eight stability. \n Possibly, too bad estimates for monodromy matrix.";
         
         /*
         //  This is only for test purposes - makes stability test for point-size set.
         //  If out estitates for this small set are not enough, 
         //  than for a set with non zero diameter can be only worse.
         {
            NBodyProblem nbody2(numberOfBodies, nDiv, Interval(M_2PI), 20, 1./128., 20, 1./128.);
            IntervalMatrix dP2(8, 8);
            dP2 = nbody2.calculateDiff(x0);
            sout  <<"\n\n dP2 = " << dP2;
            bool isStable = stability(dP2);
         }
         */
         sout << "\n\n Time used : " << timeMeter << ".\n";
         return 0;
     }
     else
     {  
         sout << "\n\n PROOF FAILED! \n"; 
         sout << "\n\n Time used : " << timeMeter << ".\n";
         return 2;
     }
     
     
   }
   catch(std::exception &e)
   {
      sout << e.what();
      return 1;
   }
}

/// @}
