/////////////////////////////////////////////////////////////////////////////
//   Package:          Choreographies in the N-Body Problem
/// @addtogroup NBodyProblem
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file nBodyProblem2.h
///
/// Definition of the class NBodyProblem2 
///
/// NBodyProblem2 class in opposite to NBodyProblem class 
/// when computing Poincare Map uses various parameters (time steps and orders)
/// from given domain and then takes intersections of images to decrease size 
/// of an upperapproximation. 
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Tomasz Kapela 2006. 
//



#ifndef _NBODYPROBLEM2_H_
#define _NBODYPROBLEM2_H_

#include "typedefs.h"
#include "nbody.h"
#include "textfile.h"
#include "mapping.h"

//////////////////////////////////////////////////////////////////////////////
//   NBodyProblem2
///
///   A brief class description.
///   A more detailed class description.
///
///   @remarks Remarks.
///
///   @warning Warnings.
///
///   @author Tomasz Kapela     @date 22-08-2006
//////////////////////////////////////////////////////////////////////////////

class NBodyProblem2 : public capd::newton::Mapping<IntervalMatrix>
{
  public:

    NBodyProblem2(int A_nBodies, int A_nDiv, Interval A_time,
                  int A_orderBegin, int A_orderEnd, int A_orderStep,
                  double A_stepBegin, double A_stepEnd, double A_stepFactor,
                  int A_XorderBegin, int A_XorderEnd, int A_XorderStep,
                  double A_XstepBegin, double A_XstepEnd, double A_XstepFactor
                 );

    virtual ~NBodyProblem2() {}

    IntervalMatrix operator[] (IntervalVector X);  // Derivative of the Map
    IntervalVector operator() (IntervalVector X);  // Value of the Map
    IntervalVector operator() (IntervalVector X, IntervalMatrix &dF);  // Value of the Map and its derivative dF

    int dimension();
    bool uniquenessTest();


    int nBodies;
    int nDiv;

   protected:
    Interval time;
    int orderBegin, orderEnd, orderStep,
        XorderBegin, XorderEnd, XorderStep;
    double stepBegin, stepEnd, stepFactor,
           XstepBegin, XstepEnd, XstepFactor;
    TimeMap timeMap;
    IntervalMatrix  E,
                    Shift,
                    R;
    IntervalVector setZC;        // Set Z x C for uniqueness test

    IntervalVector calculatePM(IntervalVector X, IntervalMatrix &dP,
                               int orderBegin, int orderEnd, int orderStep,
                               double stepBegin, double stepEnd, double stepFactor);
    IntervalMatrix generateE();
    IntervalMatrix generateShift();
    IntervalMatrix generateR();
};



inline NBodyProblem2::NBodyProblem2(int A_nBodies, int A_nDiv, Interval A_time,
                                    int A_orderBegin, int A_orderEnd, int A_orderStep,
                                    double A_stepBegin, double A_stepEnd, double A_stepFactor,
                                    int A_XorderBegin, int A_XorderEnd, int A_XorderStep,
                                    double A_XstepBegin, double A_XstepEnd, double A_XstepFactor
                                    )
                           : nBodies(A_nBodies),
                             nDiv(A_nDiv),
                             time(A_time),
                             orderBegin(A_orderBegin),  orderEnd(A_orderEnd),  orderStep(A_orderStep),
                             XorderBegin(A_XorderBegin), XorderEnd(A_XorderEnd), XorderStep(A_XorderStep),
                             stepBegin(A_stepBegin), stepEnd(A_stepEnd), stepFactor(A_stepFactor),
                             XstepBegin(A_XstepBegin), XstepEnd(A_XstepEnd), XstepFactor(A_XstepFactor),
                             timeMap(reducedNBodyEquation2D(A_nBodies),A_orderBegin, A_stepBegin)
{
   E = generateE();
   Shift = generateShift();
   R = generateR();
}



inline int NBodyProblem2::dimension()
{
   return ((nBodies-1)*4*nDiv-2);
}

// Generates matrix of expansion function
// simply, this function adds two coordinates set to zero
//   (second and third coordinate)
//

inline IntervalMatrix NBodyProblem2::generateE()
{
   IntervalMatrix E((nBodies-1)*4*nDiv,(nBodies-1)*4*nDiv-2);
   E.clear();
   E[0][0]=Interval(1.);
   for(int i=1; i<(nBodies-1)*4*nDiv-2; ++i)
      E[i+2][i]=Interval(1.);
   return E;
}


inline IntervalMatrix NBodyProblem2::generateR()
{
   IntervalMatrix R((nBodies-1)*4*nDiv-2,(nBodies-1)*4*nDiv);
   R.clear();
   R[0][0]=Interval(1.);
   for(int i=1; i<(nBodies-1)*4*nDiv-2; i++)
      R[i][i+2]=Interval(1.);
   return R;
}

inline IntervalMatrix NBodyProblem2::generateShift()
{
   IntervalMatrix S((nBodies-1)*4*nDiv,(nBodies-1)*4*nDiv);
   S.clear();
   int tupleLength = (nBodies-1)*4;
   for(int i=(nDiv-1)*tupleLength, j=0; i< nDiv*tupleLength; i++, j=(j+1)%4)
     S[j][i] = Interval(-1.);

   for(int i=(nDiv-1)*tupleLength, j=4; i< nDiv*tupleLength-4; i++, j++)
     S[j][i] = Interval(1.);

   for(int i=0; i< (nDiv-1)*tupleLength; i++)
     S[tupleLength+i][i] = Interval(1.);

   return S;
}

#endif

/// @}

